/*
    Scanbot SDK Flutter Plugin

    Copyright (c) 2019 doo GmbH. All rights reserved.
 */
package io.scanbot.sdk.flutter

import android.content.Context
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.Matrix
import android.graphics.PointF
import android.net.Uri
import com.google.gson.Gson
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel.Result
import io.scanbot.pdf.model.PageDirection
import io.scanbot.pdf.model.PageSize
import io.scanbot.pdf.model.PdfAttributes
import io.scanbot.pdf.model.PdfConfig
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.entity.Language
import io.scanbot.sdk.flutter.utils.StorageUtils
import io.scanbot.sdk.ocr.OpticalCharacterRecognizer
import io.scanbot.sdk.persistence.Page
import io.scanbot.sdk.persistence.PageFileStorage
import io.scanbot.sdk.persistence.PageStorageProcessor
import io.scanbot.sdk.persistence.PageStorageProcessor.Configuration.Size
import io.scanbot.sdk.process.ImageFilterType
import io.scanbot.sdk.tiff.model.TIFFImageWriterCompressionOptions
import io.scanbot.sdk.tiff.model.TIFFImageWriterParameters
import io.scanbot.sdk.tiff.model.TIFFImageWriterUserDefinedField
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.async
import kotlinx.coroutines.awaitAll
import kotlinx.coroutines.withContext
import org.json.JSONObject
import java.io.File
import java.util.*
import kotlin.math.max
import kotlin.math.min

class PageProcessorPlugin(
    val context: Context,
    coroutineResultHandler: CoroutineResultHandler,
    val scanbotSDK: ScanbotSDK,
    val isEncryptionEnabled: Boolean,
    val gson: Gson
) : ISdkSubplugin,
    ICoroutineActivityResultHandler by coroutineResultHandler {

    private val pageProcessor = scanbotSDK.createPageProcessor()
    private val ocrRecognizer = scanbotSDK.createOcrRecognizer()
    private val pageFileStorage = scanbotSDK.createPageFileStorage()
    override suspend fun onMethodCall(call: MethodCall, result: Result): Boolean {
        when (call.method) {
            "applyImageFilter" -> applyImageFilter(call, result)
            "rotatePageClockwise" -> rotatePageClockwise(call, result)
            "cropAndRotatePage" -> cropAndRotatePage(call, result)
            "performOcr" -> performOcr(context, call, result)
            "getOcrConfig" -> getOcrConfigs(call, result)
            "detectDocument" -> detectDocument(call, result)
            "createPage" -> createPage(context, call, result)
            "createTiff" -> createTiff(context, call, result)
            "createPdf" -> createPdf(context, call, result)
            "refreshImageUris" -> refreshImageUris(context, call, result)
            "createPageWithImageBytes" -> createPageWithImageBytes(context, call, result)
            else -> return false
        }
        return true
    }

    private suspend fun applyImageFilter(
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val pageJson = call.argument<Json>("page") ?: Json()
                val imageFilterType = call.argument<String>("filter") ?: ""
                val pageFromJson = pageJson.toPage(gson)

                val page = pageProcessor.applyFilter(
                    pageFromJson,
                    ImageFilterType.valueOf(imageFilterType)
                )
                val json = page.toJson(gson)
                insertPageUris(scanbotSDK, this, page.pageId, json)
                withContext(Dispatchers.Main) { methodResult.success(json.toString()) }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private suspend fun rotatePageClockwise(
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val pageJson =
                    call.argument<Json>("page") ?: Json()
                val times =
                    call.argument<Int>("rotationsCount") ?: 1

                val pageFromJson = pageJson.toPage(gson)

                val page = pageProcessor.rotate(
                    pageFromJson,
                    times
                )
                val json = page.toJson(gson) as JsonObject
                insertPageUris(scanbotSDK, this, page.pageId, json)

                withContext(Dispatchers.Main) {
                    methodResult.success(json.toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }


    private suspend fun cropAndRotatePage(
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val pageJson =
                    call.argument<Json>("page") ?: Json()
                val times =
                    call.argument<Int>("rotationsCount") ?: 1

                val pageFromJson = pageJson.toPage(gson)
                val polygonJson = call.argument<PolygonJson>("polygon") ?: JsonArray()
                val polygon =
                    polygonJson.map {
                        gson.fromJson(
                            JSONObject(it as Map<*, *>).toString(),
                            PointF::class.java
                        )
                    }
                        .toList()
                val page = pageProcessor.cropAndRotate(
                    pageFromJson,
                    times,
                    polygon
                )
                val json = page.toJson(gson)
                insertPageUris(scanbotSDK, this, page.pageId, json)

                withContext(Dispatchers.Main) {
                    methodResult.success(json.toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private suspend fun detectDocument(
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val pageJson =
                    call.argument<Json>("page") ?: Json()
                var page = pageJson.toPage(gson)

                // run auto document detection on it:
                page = pageProcessor.detectDocument(page)

                val json = page.toJson(gson)
                insertPageUris(scanbotSDK, this, page.pageId, json)
                withContext(Dispatchers.Main) {
                    methodResult.success(json.toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private suspend fun createPage(
        activity: Context,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val originalImageFileUri = call.argument<String>("originalImageFileUri") ?: ""
                val shouldDetectDocument = call.argument<Boolean>("shouldDetectDocument") ?: false
                val originalImageSizeLimit =
                    call.argument<Json>("originalImageSizeLimit")?.toPageProcessorSizeConfig(gson)
                        ?: Size()

                val documentImageSizeLimit =
                    call.argument<Json>("documentImageSizeLimit")?.toPageProcessorSizeConfig(gson)
                        ?: Size()


                val originalBitmap = prepareImageForImport(
                    Uri.parse(originalImageFileUri),
                    activity,
                    fixOrientation = true
                )

                val downscale = calculateDownscaleRatio(originalImageSizeLimit, originalBitmap)

                val page = pageFileStorage.add(
                    originalBitmap,
                    PageStorageProcessor.Configuration(
                        imageScale = downscale,
                        documentImageSizeLimit = documentImageSizeLimit,
                        createDocument = shouldDetectDocument
                    )
                )

                if (!shouldDetectDocument) {
                    var bitmap =
                        pageFileStorage.getImage(page.pageId, PageFileStorage.PageFileType.ORIGINAL)
                    if (bitmap != null) {
                        val downscale = calculateDownscaleRatio(documentImageSizeLimit, bitmap)
                        if (downscale < 1.0) {
                            val width = bitmap.width
                            val height = bitmap.height
                            val matrix = Matrix()
                            matrix.postScale(downscale, downscale)
                            bitmap = Bitmap.createBitmap(
                                bitmap, 0, 0,
                                width, height, matrix, false
                            )!!
                        }
                        pageFileStorage.setImageForId(
                            bitmap,
                            page.pageId,
                            PageFileStorage.PageFileType.DOCUMENT
                        )
                    }
                }

                val json = page.toJson(gson)
                insertPageUris(scanbotSDK, this, page.pageId, json)
                withContext(Dispatchers.Main) {
                    methodResult.success(json.toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private fun calculateDownscaleRatio(
        originalImageSizeLimit: Size,
        originalBitmap: Bitmap
    ): Float {
        val maxConstraintSize =
            max(originalImageSizeLimit.width, originalImageSizeLimit.height)
        val maxBitmapSize = max(originalBitmap.width, originalBitmap.height)
        val scaleTo = maxConstraintSize.toFloat() / maxBitmapSize.toFloat()
        return min(1f, scaleTo)
    }

    private suspend fun createPageWithImageBytes(
        activity: Context,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val originalImageBytes =
                    call.argument<ByteArray>("originalImageBytes") ?: byteArrayOf()
                val shouldDetectDocument = call.argument<Boolean>("shouldDetectDocument") ?: false

                val originalImageSizeLimit =
                    call.argument<Json>("originalImageSizeLimit")?.toPageProcessorSizeConfig(gson)
                        ?: Size()

                val documentImageSizeLimit =
                    call.argument<Json>("documentImageSizeLimit")?.toPageProcessorSizeConfig(gson)
                        ?: Size()


                val options: BitmapFactory.Options = BitmapFactory.Options()
                val originalBitmap = BitmapFactory.decodeByteArray(
                    originalImageBytes,
                    0,
                    originalImageBytes.size,
                    options
                )

                val downscale = calculateDownscaleRatio(originalImageSizeLimit, originalBitmap)

                val page = pageFileStorage.add(
                    originalBitmap,
                    PageStorageProcessor.Configuration(
                        imageScale = downscale,
                        documentImageSizeLimit = documentImageSizeLimit,
                        createDocument = shouldDetectDocument
                    )
                )

                if (!shouldDetectDocument) {
                    var bitmap =
                        pageFileStorage.getImage(page.pageId, PageFileStorage.PageFileType.ORIGINAL)
                    if (bitmap != null) {
                        val downscale = calculateDownscaleRatio(documentImageSizeLimit, bitmap)
                        if (downscale < 1.0) {
                            val matrix = Matrix()
                            matrix.postScale(downscale, downscale)
                            bitmap = Bitmap.createBitmap(
                                bitmap,
                                0,
                                0,
                                bitmap.width,
                                bitmap.height,
                                matrix,
                                false
                            )!!
                        }
                        pageFileStorage.setImageForId(
                            bitmap,
                            page.pageId,
                            PageFileStorage.PageFileType.DOCUMENT
                        )
                    }
                }

                val json = page.toJson(gson)
                insertPageUris(scanbotSDK, this, page.pageId, json)
                withContext(Dispatchers.Main) {
                    methodResult.success(json.toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private suspend fun getOcrConfigs(
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val configJson = JsonObject()
                val languagesJson = JsonArray()
                ocrRecognizer.getInstalledLanguages().map {
                    it.languageTags.firstOrNull() ?: it.ocrBlobLanguageTag
                }.forEach {
                    languagesJson.add(it)
                }
                configJson.addProperty("languageDataPath", ocrRecognizer.getLanguageDataPath().path)
                configJson.add("installedLanguages", languagesJson)
                withContext(Dispatchers.Main) {
                    methodResult.success(configJson.toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private suspend fun performOcr(
        context: Context,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val pagesJson = call.argument<PagesJson>("pages")!!
                val options = call.argument<Json>("options")

                val languages = (options?.get("languages") as LanguagesJson).mapNotNull {
                    Language.languageByLanguageTag(it)
                }.toSet()

                val ocrEngine = gson.fromJson(
                    options?.get("engine") as String,
                    OpticalCharacterRecognizer.EngineMode::class.java
                )
                val shouldGeneratePdf = options?.get("shouldGeneratePdf") as? Boolean
                val pdfOption = options?.get("pdfOptions")
                val configJson = JSONObject(pdfOption as Map<*, *>)
                val direction = gson.fromJson(
                    configJson["pageDirection"].toString(),
                    PageDirection::class.java
                )
                val pageSize = gson.fromJson(
                    configJson["pageSize"].toString(),
                    PageSize::class.java
                )
                val pafAttr = gson.fromJson(
                    configJson["pdfAttributes"].toString(),
                    PdfAttributes::class.java
                )
                val pages =
                    pagesJson.map {
                        gson.fromJson(
                            JSONObject(it as Map<*, *>).toString(),
                            Page::class.java
                        )
                    }.toList()

                ocrRecognizer.setOcrConfig(
                    OpticalCharacterRecognizer.OcrConfig(
                        engineMode = ocrEngine,
                        languages = languages
                    )
                )
                val ocrResult = if (shouldGeneratePdf == true) {
                    ocrRecognizer.recognizeTextWithPdfFromPages(
                        pages, PdfConfig.defaultConfig().copy(
                            pageDirection = direction,
                            pageSize = pageSize,
                            pdfAttributes = pafAttr
                        )
                    )
                } else {
                    ocrRecognizer.recognizeTextFromPages(pages)
                }
                val tmpFile = ocrResult?.sandwichedPdfDocumentFile
                val newFile = tmpFile?.let {
                    val newFile = movePdfFileToStorage(it, context)
                    newFile
                }

                val ocrResultJson = ocrResultToJson(ocrResult, newFile)
                withContext(Dispatchers.Main) {
                    methodResult.success(ocrResultJson.toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private suspend fun createTiff(
        activity: Context,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val pagesJson = call.argument<PagesJson>("pages")!!
                val options = call.argument<Json>("options")
                val binarized = options?.get("binarized") as Boolean
                val compressionString = (options?.get("compression") as? String)
                    .takeUnless { it.isNullOrBlank() }
                    ?: "NONE"

                val imageFiles =
                    pagesJson.map {
                        gson.fromJson(
                            JSONObject(it as Map<*, *>).toString(),
                            Page::class.java
                        )
                    }
                        .map {
                            async {
                                pageFileStorage.getImageURI(
                                    it.pageId,
                                    PageFileStorage.PageFileType.DOCUMENT
                                )
                            }
                        }.awaitAll().map {
                            File(it.path)
                        }


                val dpi = (options?.get("dpi") as? Int) ?: 200

                val compression =
                    TIFFImageWriterCompressionOptions.valueOf("COMPRESSION_$compressionString")


                // no custom fields supported yet.
                val customFields = ArrayList<TIFFImageWriterUserDefinedField>()

                val params = TIFFImageWriterParameters(
                    if (binarized) ImageFilterType.PURE_BINARIZED else ImageFilterType.NONE,
                    dpi,
                    compression,
                    customFields
                )

                val tiffTargetFile = StorageUtils.generateRandomPluginStorageFile("tiff", activity)
                val tiffWriter = scanbotSDK.createTiffWriter()
                tiffWriter.writeTIFFFromFiles(
                    imageFiles,
                    isEncryptionEnabled,
                    tiffTargetFile,
                    params
                )

                withContext(Dispatchers.Main) {
                    methodResult.success(Uri.fromFile(tiffTargetFile).toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private suspend fun createPdf(
        context: Context,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val pagesJson = call.argument<PagesJson>("pages")!!
                val options = call.argument<Json>("options")
                val configJson = JSONObject(options as Map<*, *>)
                val direction = gson.fromJson(
                    configJson["pageDirection"].toString(),
                    PageDirection::class.java
                )
                val pageSize = gson.fromJson(
                    configJson["pageSize"].toString(),
                    PageSize::class.java
                )
                val pafAttr = gson.fromJson(
                    configJson["pdfAttributes"].toString(),
                    PdfAttributes::class.java
                )
                val pages = pagesJson.map {
                    gson.fromJson(
                        JSONObject(it as Map<*, *>).toString(),
                        Page::class.java
                    )
                }
                val pdfRenderer = scanbotSDK.createPdfRenderer()

                val tmpfile = pdfRenderer.renderDocumentFromPages(
                    pages,
                    PdfConfig.defaultConfig().copy(
                        pageDirection = direction,
                        pageSize = pageSize,
                        pdfAttributes = pafAttr
                    )
                )
                val file = movePdfFileToStorage(tmpfile, context)
                withContext(Dispatchers.Main) {
                    methodResult.success(Uri.fromFile(file).toString())
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private suspend fun movePdfFileToStorage(tempPdfFile: File?, context: Context): File {
        try {
            val pdfOutputUri =
                Uri.fromFile(StorageUtils.generateRandomPluginStorageFile("pdf", context))
            val newFile = File(pdfOutputUri.path)
            tempPdfFile?.let { StorageUtils.copyFile(it, newFile) }
            return newFile
        } catch (e: Throwable) {
            throw e
        } finally {
            tempPdfFile?.let { StorageUtils.forceDelete(it) }
        }
    }

    private suspend fun refreshImageUris(context: Context, call: MethodCall, methodResult: Result) {
        // This method is actually not required on Android. We just provide a dummy implementation
        // to stay consistent with the Flutter API and iOS. See the comments of the iOS implementation.
        withContext(Dispatchers.Default) {
            try {
                val pagesJson = call.argument<PagesJson>("pages")!!
                val array = JsonArray()
                pagesJson.onEach { pageJson ->
                    val page = pageJson.toPage(gson)
                    val newPageJson = page.toJson(gson)
                    insertPageUris(scanbotSDK, this, page.pageId, newPageJson)
                    array.add(newPageJson)
                }

                val resultJson = JsonObject()
                resultJson.add("pages", array)
                withContext(Dispatchers.Main) { methodResult.success(resultJson.toString()) }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        e::class.java.simpleName,
                        e.message,
                        null
                    )
                }
            }
        }
    }
}

