package io.scanbot.sdk.flutter.camera.detectorWrappers

import android.content.pm.PackageManager
import android.graphics.Color
import android.util.Log
import android.view.View.OnLayoutChangeListener
import android.widget.FrameLayout
import androidx.camera.view.PreviewView
import androidx.core.graphics.alpha
import androidx.core.graphics.blue
import androidx.core.graphics.green
import androidx.core.graphics.red
import com.google.gson.Gson
import com.google.gson.JsonObject
import io.flutter.plugin.common.MethodChannel
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.barcode.BarcodeDetectorFrameHandler
import io.scanbot.sdk.barcode.ScanbotBarcodeDetector
import io.scanbot.sdk.barcode.entity.BarcodeItem
import io.scanbot.sdk.barcode.entity.BarcodeScanningResult
import io.scanbot.sdk.barcode.ui.BarcodePolygonsView
import io.scanbot.sdk.barcode.ui.BarcodeScannerView
import io.scanbot.sdk.barcode.ui.IBarcodeScannerViewCallback
import io.scanbot.sdk.camera.CaptureInfo
import io.scanbot.sdk.camera.FrameHandlerResult
import io.scanbot.sdk.flutter.*
import io.scanbot.sdk.flutter.camera.CustomViewDisplayOwner
import io.scanbot.sdk.flutter.camera.config.BarcodeScannerConfiguration
import io.scanbot.sdk.flutter.camera.usecase.SaveBarcodePreviewFrameUseCase
import io.scanbot.sdk.flutter.camera.usecase.SaveBarcodeTakenPictureUseCase
import io.scanbot.sdk.flutter.utils.convertDpToPx
import io.scanbot.sdk.flutter.utils.convertPxToDp
import io.scanbot.sdk.ui.camera.CameraUiSettings
import io.scanbot.sdk.ui.camera.ScanbotCameraContainerView
import io.scanbot.sdk.ui.camera.ScanbotCameraXView
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeImageGenerationType
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeImageGenerationType.*
import io.scanbot.sdk.ui.view.barcode.configuration.json.toSdk
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.single
import java.util.*
import io.scanbot.sdk.flutter.utils.LifecycleOwnerWrapper

class BarcodeDetector(
    private val wrapper: LifecycleOwnerWrapper<BarcodeScannerView>,
    private val coroutineScope: CoroutineScope,
    private val cameraChannel: MethodChannel,
    private val detectorChannel: MethodChannel,
    private val finderChannel: MethodChannel,
    override val scanbotSDK: ScanbotSDK,
    private val customViewDisplayOwner: CustomViewDisplayOwner,
    private val creationParams: HashMap<String, Any>,
    override val gson: Gson
) : IDetectorWrapper, CoroutineScope by coroutineScope {
    private val returnContext = Dispatchers.Main
    private val barcodeFileStorage = scanbotSDK.createBarcodeFileStorage()
    private val imageFileIOProcessor = scanbotSDK.fileIOProcessor()
    private val scannerView: BarcodeScannerView = wrapper.view
    private val saveBarcodePreviewFrameUseCase =
        SaveBarcodePreviewFrameUseCase(barcodeFileStorage, imageFileIOProcessor)
    private val saveBarcodeTakenPictureUseCase =
        SaveBarcodeTakenPictureUseCase(barcodeFileStorage, imageFileIOProcessor)

    private var cachedBarcodeResult: BarcodeScanningResult? = null
    private var barcodeImageGenerationType: BarcodeImageGenerationType =
        BarcodeImageGenerationType.NONE
    private var isSavingBarcodeSnapshot = false
    private var shouldEnableDetection = true
    private var automaticSelectionEnabled = true

    private val handler = BarcodeDetectorFrameHandler.ResultHandler { result ->
        runBlocking {
            when (result) {
                is FrameHandlerResult.Success -> {
                    if (!automaticSelectionEnabled) {
                        return@runBlocking
                    }
                    result.value?.let { result ->
                        try {
                            val json = gson.toJsonTree(result).asJsonObject
                            when (barcodeImageGenerationType) {
                                NONE -> {
                                    withContext(returnContext) {
                                        scannerView.viewController.isFrameProcessingEnabled = false
                                        callMethodSync<Any>(
                                            detectorChannel,
                                            "onBarcodeDetected",
                                            json
                                        )
                                    }
                                    scannerView.viewController.isFrameProcessingEnabled =
                                        shouldEnableDetection
                                }

                                VIDEO_FRAME -> {
                                    scannerView.viewController.isFrameProcessingEnabled = false
                                    withContext(Dispatchers.Default) {
                                        savePreviewFrameAndReturn(
                                            result,
                                            json
                                        )
                                    }
                                    withContext(returnContext) {
                                        callMethodSync<Any>(
                                            detectorChannel,
                                            "onBarcodeDetected",
                                            json
                                        )
                                    }
                                    scannerView.viewController.isFrameProcessingEnabled =
                                        shouldEnableDetection
                                }

                                CAPTURED_IMAGE -> {
                                    if (!isSavingBarcodeSnapshot) {
                                        cachedBarcodeResult = result
                                    }
                                }
                            }
                        } catch (e: Throwable) {
                            e.printStackTrace()
                        }
                    } ?: run { cachedBarcodeResult == null }
                }

                is FrameHandlerResult.Failure -> {
                    withContext(returnContext) {
                        val json = scanbotSDK.getLicenseStatusJson()
                        detectorChannel.invokeMethod(
                            "onLicenseError",
                            json.toString()
                        )
                    }
                }
            }
        }
        false
    }

    override fun attach(arguments: Json) {
        creationParams.apply {
            scannerView.initCamera(CameraUiSettings(true))
            val cameraXView =
                scannerView.findViewById<ScanbotCameraContainerView>(R.id.scanbot_camera_view)
            val previewView =
                cameraXView.findViewById<PreviewView>(io.scanbot.sdk.base.R.id.camera_preview_view)
            previewView.implementationMode = PreviewView.ImplementationMode.COMPATIBLE
            (cameraXView.getChildAt(0) as ScanbotCameraXView).lifecycleOwner =
                customViewDisplayOwner
            configureFinderView()
            configureSelectionOverlay()
            val barcodeScannerConfiguration = (arguments).toBarcodeScannerConfig(gson)
            val barcodeDetector = scanbotSDK.createBarcodeDetector()

            scannerView.viewController.barcodeDetectionInterval = 0L

            updateScannerConfiguration(barcodeDetector, barcodeScannerConfiguration)

            addDetectorMethodHandler()
            scannerView.initDetectionBehavior(
                barcodeDetector,
                handler,
                object : IBarcodeScannerViewCallback {
                    override fun onSelectionOverlayBarcodeClicked(barcodeItem: BarcodeItem) {
                        if (!automaticSelectionEnabled) {
                            val json = gson.toJsonTree(barcodeItem).asJsonObject
                            cameraChannel.invokeMethod("onBarcodeSelected", gson.toJson(json))
                        }
                    }

                    override fun onCameraOpen() {
                        creationParams.apply {
                            launch {
                                withContext(returnContext) {
                                    checkParam<Boolean>("flashEnabled")?.let {
                                        scannerView.viewController.useFlash(it)
                                    }
                                    checkParam<Float>("cameraZoomFactor")?.let {

                                        cameraXView.setPhysicalZoom(it * (12 - 1) + 1)
                                    }

                                    cameraChannel.invokeMethod("onCameraOpen", "")
                                }
                            }
                        }
                    }

                    override fun onPictureTaken(image: ByteArray, captureInfo: CaptureInfo) {
                        //todo possibly need to fix threading
                        launch {
                            isSavingBarcodeSnapshot = true
                            saveSnappedImageAndReturn(image, captureInfo)
                            withContext(Dispatchers.Main) { scannerView.viewController.startPreview() }
                            isSavingBarcodeSnapshot = false
                        }
                    }
                })

            barcodeImageGenerationType = barcodeScannerConfiguration.barcodeImageGenerationType
            when (barcodeScannerConfiguration.barcodeImageGenerationType) {
                VIDEO_FRAME -> barcodeDetector.modifyConfig { this.setSaveCameraPreviewFrame(true) }
                CAPTURED_IMAGE -> {
                    scannerView.viewController.autoSnappingEnabled = true
                    barcodeScannerConfiguration.autoSnappingSensitivity?.let {
                        scannerView.viewController.setAutoSnappingSensitivity(
                            it
                        )
                    }
                }

                else -> {
                }
            }
        }
    }

    private suspend fun saveSnappedImageAndReturn(image: ByteArray, captureInfo: CaptureInfo) {
        try {
            val json = cachedBarcodeResult?.let { gson.toJsonTree(it).asJsonObject } ?: JsonObject()
            withContext(returnContext) { cameraChannel.invokeMethod("onShowHeavyProcessing", true) }
            val filePath =
                saveBarcodeTakenPictureUseCase.saveTakenPicture(image, captureInfo.imageOrientation)
                    .single()
            json.addProperty("barcodeImageURI", filePath)
            withContext(returnContext) {
                cameraChannel.invokeMethod(
                    "onShowHeavyProcessing",
                    false
                )
            }
            withContext(returnContext) {
                callMethodSync<Any>(
                    detectorChannel,
                    "onBarcodeDetected",
                    json
                )
            }
            scannerView.viewController.isFrameProcessingEnabled = shouldEnableDetection
        } catch (e: Throwable) {
            Log.e("ScanbotCameraModule", "error while saving snapped image", e)
        }
    }

    private suspend fun savePreviewFrameAndReturn(result: BarcodeScanningResult, json: JsonObject) {
        result.previewFrame?.let {
            withContext(returnContext) { cameraChannel.invokeMethod("onShowHeavyProcessing", true) }
            val filePath = saveBarcodePreviewFrameUseCase.savePreviewFrame(it).single()
            json.addProperty("barcodeImageURI", filePath)
            withContext(returnContext) {
                cameraChannel.invokeMethod(
                    "onShowHeavyProcessing",
                    false
                )
            }
        }
    }

    private fun updateScannerConfiguration(
        barcodeDetector: ScanbotBarcodeDetector,
        barcodeScannerConfiguration: BarcodeScannerConfiguration
    ) {
        barcodeDetector.modifyConfig {
            barcodeScannerConfiguration.acceptedDocumentFormats?.let {
                this.setAcceptedDocumentFormats(it)
            }
            barcodeScannerConfiguration.barcodeFormats?.let { this.setBarcodeFormats(it) }
            barcodeScannerConfiguration.engineMode?.let { this.setEngineMode(it) }
            barcodeScannerConfiguration.additionalParameters?.let { params ->
                this.modifyAdditionalConfig {
                    params.enableGS1Decoding?.let { this.setGs1DecodingEnabled(it) }
                    params.maximumTextLength?.let { this.setMaximumTextLength(it) }
                    params.stripCheckDigits?.let { this.setStripCheckDigits(it) }
                    params.minimumTextLength?.let { this.setMinimumTextLength(it) }

                    params.minimum1DBarcodesQuietZone?.let { this.setMinimum1DQuietZoneSize(it) }
                    params.msiPlesseyChecksumAlgorithm?.let {
                        this.setMsiPlesseyChecksumAlgorithms(
                            EnumSet.of(it.toSdk())
                        )
                    }
                }
            }
        }
    }

    private var frameLayout: FrameLayout? = null

    private val finderListener =
        OnLayoutChangeListener { view, left, top, right, bottom, _, _, _, _ ->
            frameLayout?.notifyFlutterFinder(
                left, top, right, bottom
            )
        }

    private fun FrameLayout.registerFinderListener() {
        frameLayout = this
        addOnLayoutChangeListener(finderListener)
    }

    private fun FrameLayout.notifyFlutterFinder(
        left: Int,
        top: Int,
        right: Int,
        bottom: Int
    ) {
        finderChannel.invokeMethod(
            "finderSizeChanged", mapOf(
                "left" to convertPxToDp(context, left.toFloat()),
                "top" to convertPxToDp(context, top.toFloat()),
                "right" to convertPxToDp(context, right.toFloat()),
                "bottom" to convertPxToDp(context, bottom.toFloat()),
            )
        )
    }

    private fun Json.configureFinderView() {
        val context = scannerView.context
        val finderViewController = scannerView.finderViewController
        checkParam<Json>("finder") { it ->
            it.toFinderConfig(gson).also { config ->
                finderViewController.setStrokeColor(Color.TRANSPARENT)
                finderViewController.setOverlayColor(Color.TRANSPARENT)
                config.finderAspectRatio?.let {
                    finderViewController.setRequiredAspectRatios(listOf(it))
                }
                config.finderInsets?.let { inset ->
                    finderViewController.setFinderInset(
                        inset.left?.let { it1 -> convertDpToPx(context, it1).toInt() },
                        inset.top?.let { it1 -> convertDpToPx(context, it1).toInt() },
                        inset.right?.let { it1 -> convertDpToPx(context, it1).toInt() },
                        inset.bottom?.let { it1 -> convertDpToPx(context, it1).toInt() },
                    )
                }

                finderViewController.addFinderPlaceholder(
                    FrameLayout(scannerView.context).apply {
                        registerFinderListener()
                    })
            }
            finderViewController.setFinderEnabled(true)
        } ?: finderViewController.setFinderEnabled(false)
    }

    /** This method called to set values from Json to [BarcodeScannerView.selectionOverlayController]. */
    private fun Json.configureSelectionOverlay() {
        val selectionOverlayController = scannerView.selectionOverlayController
        checkParam<Json>("overlayConfiguration") { it ->
            it.toSelectionOverlayConfig(gson).also { config ->
                config.overlayEnabled.let {
                    selectionOverlayController.setEnabled(it)
                }
                config.automaticSelectionEnabled.let {
                    automaticSelectionEnabled = it
                }

                selectionOverlayController.setBarcodeAppearanceDelegate(object :
                    BarcodePolygonsView.BarcodeAppearanceDelegate {
                    override fun getPolygonStyle(
                        defaultStyle: BarcodePolygonsView.BarcodePolygonStyle,
                        barcodeItem: BarcodeItem,
                    ): BarcodePolygonsView.BarcodePolygonStyle {
                        val polygonColor: Int = config.polygonColor?.let { Color.parseColor(it) }
                            ?: defaultStyle.fillColor

                        return defaultStyle.copy(
                            strokeColor = polygonColor,
                            fillColor = polygonColor.makeTransparent(),
                        )
                    }

                    override fun getTextViewStyle(
                        defaultStyle: BarcodePolygonsView.BarcodeTextViewStyle,
                        barcodeItem: BarcodeItem
                    ): BarcodePolygonsView.BarcodeTextViewStyle {
                        val textColor =
                            config.textColor?.let { Color.parseColor(it) } ?: defaultStyle.textColor
                        val textContainerColor =
                            config.textContainerColor?.let { Color.parseColor(it) }
                                ?: defaultStyle.textContainerColor

                        return defaultStyle.copy(
                            textColor = textColor,
                            textContainerColor = textContainerColor,
                            textFormat = config.textFormat,
                        )
                    }
                })
            }
        }
    }

    private fun Int.makeTransparent(alphaMultiplier: Float = POLYGON_FILL_ALPHA): Int {
        return Color.argb((this.alpha * alphaMultiplier).toInt(), this.red, this.green, this.blue)
    }

    private fun addDetectorMethodHandler() {
        detectorChannel.setMethodCallHandler { call, _ ->
            when (call.method) {
                "disable" -> {
                    shouldEnableDetection = false
                    scannerView.viewController.isFrameProcessingEnabled = false
                }

                "enable" -> {
                    shouldEnableDetection = true
                    scannerView.viewController.isFrameProcessingEnabled = true
                }
            }
        }
    }

    override fun detach() {
        frameLayout?.removeOnLayoutChangeListener(finderListener)
        detectorChannel.setMethodCallHandler(null)
    }

    override fun toggleContinuousFocus() {
        scannerView.viewController.continuousFocus()
    }

    override fun isFlashAvailable(): Boolean {
        return scannerView.context.packageManager.hasSystemFeature(PackageManager.FEATURE_CAMERA_FLASH)
    }

    override fun flashEnabled(enabled: Boolean) {
        scannerView.viewController.useFlash(enabled)
    }

    private companion object {
        private const val POLYGON_FILL_ALPHA = 0.2f
    }
}
