package io.scanbot.sdk.flutter.camera.detectorWrappers

import android.content.pm.PackageManager
import android.graphics.Color
import android.view.View.OnLayoutChangeListener
import android.widget.FrameLayout
import androidx.camera.view.PreviewView
import androidx.core.view.isVisible
import com.google.gson.Gson
import com.google.gson.JsonObject
import io.flutter.plugin.common.MethodChannel
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.camera.CaptureInfo
import io.scanbot.sdk.camera.FrameHandlerResult
import io.scanbot.sdk.camera.PictureCallback
import io.scanbot.sdk.flutter.*
import io.scanbot.sdk.flutter.camera.CustomViewDisplayOwner
import io.scanbot.sdk.flutter.camera.config.MedicalCertificateScannerConfiguration
import io.scanbot.sdk.flutter.camera.usecase.MedicalCertificateImageStorage
import io.scanbot.sdk.flutter.camera.usecase.SaveCroppedMCUseCase
import io.scanbot.sdk.flutter.utils.convertDpToPx
import io.scanbot.sdk.flutter.utils.convertPxToDp
import io.scanbot.sdk.mcrecognizer.MedicalCertificateAutoSnappingController
import io.scanbot.sdk.mcrecognizer.MedicalCertificateFrameHandler
import io.scanbot.sdk.mcrecognizer.entity.MedicalCertificateRecognizerResult
import io.scanbot.sdk.ui.camera.FinderOverlayView
import io.scanbot.sdk.ui.camera.ScanbotCameraXView
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.single


class MedicalCertificateDetector(
    private val containerView: FrameLayout,
    private val coroutineScope: CoroutineScope,
    private val cameraChannel: MethodChannel,
    private val detectorChannel: MethodChannel,
    private val finderChannel: MethodChannel,
    override val scanbotSDK: ScanbotSDK,
    private val customViewDisplayOwner: CustomViewDisplayOwner,
    private val creationParams: HashMap<String, Any>,
    override val gson: Gson
) : IDetectorWrapper, CoroutineScope by coroutineScope {
    private val returnContext = Dispatchers.Main
    private val scannerView = containerView.findViewById<ScanbotCameraXView>(R.id.cameraView)
    private val finderView = containerView.findViewById<FinderOverlayView>(R.id.finder_overlay)
    private val medicalCertificateFrameHandler = MedicalCertificateFrameHandler.attach(
        scannerView,
        scanbotSDK.createMedicalCertificateRecognizer()
    )
    private val medicalCertificateRecognizer = scanbotSDK.createMedicalCertificateRecognizer()
    private var medicalCertificateScannerConfiguration: MedicalCertificateScannerConfiguration? =
        null
    private lateinit var savingJob: Job
    private var shouldEnableDetection = true

    private val imageFileIOProcessor = scanbotSDK.fileIOProcessor()
    val saveImageUseCase = SaveCroppedMCUseCase(
        MedicalCertificateImageStorage(containerView.context.applicationContext),
        imageFileIOProcessor
    )
    private val handler = MedicalCertificateFrameHandler.ResultHandler { result ->
        runBlocking {
            when (result) {
                is FrameHandlerResult.Success -> {
                    medicalCertificateFrameHandler.isEnabled = false
                    result.value?.let { result ->
                        try {
                            val json =
                                if (medicalCertificateScannerConfiguration?.captureHighResolutionImage == true) {
                                    //return recognition sucessfull only after snap was sucessfull
                                    gson.toJsonTree(result.copy(recognitionSuccessful = false)).asJsonObject
                                } else {
                                    gson.toJsonTree(result).asJsonObject
                                }
                            withContext(returnContext) {
                                callMethodSync<Any>(
                                    detectorChannel,
                                    "onMedicalCertificateDetected",
                                    json
                                )
                            }
                            medicalCertificateFrameHandler.isEnabled = shouldEnableDetection
                        } catch (e: Throwable) {
                            e.printStackTrace()
                        }
                    }
                }

                is FrameHandlerResult.Failure -> {
                    withContext(returnContext) {
                        val json = scanbotSDK.getLicenseStatusJson()
                        detectorChannel.invokeMethod(
                            "onLicenseError",
                            json.toString()
                        )
                    }
                }
            }
        }
        false
    }

    override fun attach(arguments: Json) {
        creationParams.apply {
            medicalCertificateFrameHandler.addResultHandler(handler)
            scannerView.lifecycleOwner = customViewDisplayOwner
            val previewView =
                scannerView.findViewById<PreviewView>(io.scanbot.sdk.base.R.id.camera_preview_view)
            previewView.implementationMode = PreviewView.ImplementationMode.COMPATIBLE
            configureFinderView()
            medicalCertificateScannerConfiguration =
                (arguments).toMedicalCertificateConfig(gson)

            medicalCertificateScannerConfiguration?.let { config ->
                updateScannerConfiguration(
                    medicalCertificateFrameHandler,
                    config
                )
            }

            addDetectorMethodHandler()

            scannerView.setCameraOpenCallback {
                creationParams.apply {
                    launch {
                        withContext(returnContext) {
                            checkParam<Boolean>("flashEnabled")?.let {
                                scannerView.useFlash(it)
                            }
                            checkParam<Float>("cameraZoomFactor")?.let {
                                scannerView.setPhysicalZoom(it * (12 - 1) + 1)
                            }

                            cameraChannel.invokeMethod("onCameraOpen", "")
                        }
                    }
                }
            }
            scannerView.addPictureCallback(object : PictureCallback() {
                override fun onPictureTaken(image: ByteArray, captureInfo: CaptureInfo) {
                    //todo possibly need to fix threading
                    if (::savingJob.isInitialized) {
                        savingJob.cancel()
                    }
                    savingJob = launch(Dispatchers.Default) {
                        medicalCertificateSnapped(image, captureInfo)
                        scannerView.startPreview()
                    }
                }
            })
        }
    }

    suspend fun medicalCertificateSnapped(image: ByteArray, captureInfo: CaptureInfo) {
        medicalCertificateFrameHandler.isEnabled = false
        withContext(returnContext) {
            cameraChannel.invokeMethod(
                "onShowHeavyProcessing",
                true
            )
        }
        val mcScanResult = medicalCertificateRecognizer.recognizeMcJpeg(
            image, captureInfo.imageOrientation,
            returnCroppedDocument = true,
            recognizePatientInfo = medicalCertificateScannerConfiguration?.recognizePatientInfo
                ?: true,
            recognizeBarcode = medicalCertificateScannerConfiguration?.recognizeBarcode ?: true
        )

        val rtuResult = mcScanResult?.let {
            if (it.recognitionSuccessful) {
                it
            } else {
                MedicalCertificateRecognizerResult(false)
            }
        } ?: MedicalCertificateRecognizerResult(false)

        val json = gson.toJsonTree(rtuResult).asJsonObject

        rtuResult.croppedImage?.let {
            val image = saveImageUseCase.savePreviewFrame(it).single()
            if (image.isNotEmpty()) {
                json.addProperty("croppedDocumentURI", image)
            }
        }

        withContext(returnContext) {
            cameraChannel.invokeMethod(
                "onShowHeavyProcessing",
                false
            )
            callMethodSync<Any>(detectorChannel, "onMedicalCertificateDetected", json)
            medicalCertificateFrameHandler.isEnabled = shouldEnableDetection
        }
    }

    private fun updateScannerConfiguration(
        frameHandler: MedicalCertificateFrameHandler,
        medicalCertificateScannerConfiguration: MedicalCertificateScannerConfiguration
    ) {
        medicalCertificateScannerConfiguration?.captureHighResolutionImage?.let {
            val autoSnappingController =
                MedicalCertificateAutoSnappingController.attach(scannerView, frameHandler)
            autoSnappingController.setForceAutofocusBeforeSnap(false)
            autoSnappingController.setShakeDetectionEnabled(true)

            autoSnappingController.isEnabled = it
            frameHandler.extractDocumentImage = false

            medicalCertificateScannerConfiguration?.recognizeBarcode?.let {
                frameHandler.recognizeBarcode = it
            }
            medicalCertificateScannerConfiguration?.recognizePatientInfo?.let {
                frameHandler.recognizePatientInfo = it
            }
        }
    }

    private var frameLayout: FrameLayout? = null

    private val finderListener =
        OnLayoutChangeListener { view, left, top, right, bottom, _, _, _, _ ->
            frameLayout?.notifyFlutterFinder(
                left, top, right, bottom
            )
        }

    private fun FrameLayout.registerFinderListener() {
        frameLayout = this
        addOnLayoutChangeListener(finderListener)
    }

    private fun FrameLayout.notifyFlutterFinder(
        left: Int,
        top: Int,
        right: Int,
        bottom: Int
    ) {
        finderChannel.invokeMethod(
            "finderSizeChanged", mapOf(
                "left" to convertPxToDp(context, left.toFloat()),
                "top" to convertPxToDp(context, top.toFloat()),
                "right" to convertPxToDp(context, right.toFloat()),
                "bottom" to convertPxToDp(context, bottom.toFloat()),
            )
        )
    }

    private fun Json.configureFinderView() {
        val context = scannerView.context
        checkParam<Json>("finder") { it ->
            it.toFinderConfig(gson).also { config ->
                finderView.setStrokeColor(Color.TRANSPARENT)
                finderView.setOverlayColor(Color.TRANSPARENT)
                config.finderAspectRatio?.let {
                    finderView.setRequiredAspectRatios(listOf(it))
                }
                config.finderInsets?.let { inset ->
                    finderView.setFinderInset(
                        inset.left?.let { it1 -> convertDpToPx(context, it1).toInt() },
                        inset.top?.let { it1 -> convertDpToPx(context, it1).toInt() },
                        inset.right?.let { it1 -> convertDpToPx(context, it1).toInt() },
                        inset.bottom?.let { it1 -> convertDpToPx(context, it1).toInt() },
                    )
                }

                finderView.setFinderPlaceholder(
                    FrameLayout(scannerView.context).apply {
                        registerFinderListener()
                    })
            }
            finderView.isVisible = true
        } ?: kotlin.run { finderView.isVisible = false }
    }

    private fun addDetectorMethodHandler() {
        detectorChannel.setMethodCallHandler { call, result ->
            when (call.method) {
                "disable" -> {
                    medicalCertificateFrameHandler.isEnabled = false
                    shouldEnableDetection = false
                }

                "enable" -> {
                    medicalCertificateFrameHandler.isEnabled = true
                    shouldEnableDetection = true
                }
            }
        }
    }

    override fun detach() {
        frameLayout?.removeOnLayoutChangeListener(finderListener)
        detectorChannel.setMethodCallHandler(null)
    }

    override fun toggleContinuousFocus() {
        scannerView.continuousFocus()
    }

    override fun isFlashAvailable(): Boolean {
        return scannerView.context.packageManager.hasSystemFeature(PackageManager.FEATURE_CAMERA_FLASH)
    }

    override fun flashEnabled(enabled: Boolean) {
        scannerView.useFlash(enabled)
    }
}
