//
//  SDKCallHandler.swift
//  Runner
//
//  Created by Zhenya Knizhnik on 22.08.19.
//  Copyright © 2019 The Chromium Authors. All rights reserved.
//

import Foundation
import Flutter
import ScanbotSDK

class CallParameters: NSObject {
    let call: FlutterMethodCall
    let result: FlutterResult
    
    init(call: FlutterMethodCall, result: @escaping FlutterResult) {
        self.call = call
        self.result = result
    }
}

class CallUIParameters: CallParameters {
    let presenter: UIViewController?
    
    init(call: FlutterMethodCall, result: @escaping FlutterResult, presenter: UIViewController?) {
        self.presenter = presenter
        super.init(call: call, result: result)
    }
}

class SDKCallHandler: CallErrorHandler {
    let rtuuiPresenter = RTUUIPresenter()
    var dataFormatterChannel: FlutterMethodChannel?
    
    func handle(call: FlutterMethodCall, rootViewController: UIViewController?, result: @escaping FlutterResult) {
        
        let uiParameters = CallUIParameters(call: call, result: result, presenter: rootViewController)
        let parameters = CallParameters(call: call, result: result)
        
        switch MethodName(fromRawValue: call.method) {
        case .initScanbotSdk: self.initSDK(parameters)
        case .getLicenseStatus: self.getLicenseStatus(result: result)
        case .cleanupStorage: self.cleanupStorage(result: result)
        case .rotatePageClockwise: self.rotatePage(parameters)
        case .cropAndRotatePage: self.cropAndRotatePage(parameters)
        case .deletePage: self.deletePage(parameters)
        case .applyImageFilter: self.applyImageFilter(parameters)
        case .getFilteredDocumentPreviewUri: self.getFilteredDocumentPreviewURL(parameters)
        case .createPdf: self.createPDF(parameters)
        case .createTiff: self.createTIFF(parameters)
        case .createPage: self.createPage(parameters)
        case .createPageWithImageBytes: self.createPageWithImageBytes(parameters)
        case .detectDocument: self.detectDocument(parameters)
        case .performOcr: self.performOCR(parameters)
        case .estimateBlurOnPage: self.estimateQuality(parameters)
        case .getOcrConfig: self.getOCRConfig(result: result)
        case .refreshImageUris: self.refreshImageURLS(call: call, result: result)
        case .getDecryptedDataFromFile: self.getDecryptFileFromURL(parameters)
        case .detectBarcodesOnImages: self.detectBarcodesOnImages(parameters)
        case .startDocumentScanner: self.rtuuiPresenter.showDocumentScanner(uiParameters)
        case .startBarcodeScanner: self.rtuuiPresenter.showBarcodeScanner(uiParameters)
        case .startMrzScanner: self.rtuuiPresenter.showMRZScanner(uiParameters)
        case .startEhicScanner: self.rtuuiPresenter.showEHICScanner(uiParameters)
        case .startCroppingScreen: self.rtuuiPresenter.showCropScreen(uiParameters)
        case .startBatchBarcodeScanner:
            self.rtuuiPresenter.showBatchScanner(uiParameters, dataFormatterChannel: self.dataFormatterChannel)
        case .startLicensePlateScanner: self.rtuuiPresenter.showLicensePlateScanner(uiParameters)
        case .startGenericDocumentRecognizer: self.rtuuiPresenter.showGenericDocumentRecognizer(uiParameters)
        case .unknown : self.returnUnknownCallError(call: call, result: result)
        }
    }
    
    private func getLicenseStatus(result: @escaping FlutterResult) {
        let date = Scanbot.licenseExpirationDate()
        let isValid = Scanbot.isLicenseValid()
        let status = Scanbot.licenseStatus()
        
        let statusResult = LicenseStatus(isLicenseValid: isValid, status: status, expirationDate: date)
        result(statusResult.encodedString)
    }
    
    private func initSDK(_ parameters: CallParameters) {
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let config = SDKInitializationConfig(from: arguments) else {
            self.returnArgumentsError(call: parameters.call, result: parameters.result)
            return
        }
        ScanbotSDKInitializer.initialize(config: config, result: parameters.result)
    }
    
    private func getDecryptFileFromURL(_ parameters: CallParameters) {
        let result = parameters.result
        
        DispatchQueue.global(qos: .background).async {
            guard let arguments = parameters.call.arguments as? [String: Any],
                  let fileURLString = (arguments["fileUri"] as? String)?.removingPercentEncoding,
                  let dataToDecrypt = try? Data(contentsOf: URL(fileURLWithPath: fileURLString)) else {
                
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            // To mimic Android SDK behavior we are just returning the data of the file, in case no encrypter  is set.
            var returnData = dataToDecrypt
            
            // We are using imageStorage encryptor, as the encryptor should be the same for all cases.
            if let encrypter = ScanbotUI.defaultImageStoreEncrypter() {
                returnData = encrypter.decryptData(dataToDecrypt)
            }
            DispatchQueue.main.async {
                result(returnData)
            }
        }
    }
    
    private func cleanupStorage(result: @escaping FlutterResult) {
        SBSDKUIPageFileStorage.default().removeAll()
        
        if let url = PluginPreferences.shared.storageCustomFilesURL {
            self.clearDirectory(directoryURL: url)
        }
        
        result(nil)
    }
    
    private func clearDirectory(directoryURL: URL) {
        let fileManager = FileManager.default
        do {
            let urls = try fileManager.contentsOfDirectory(at: directoryURL,
                                                           includingPropertiesForKeys: nil,
                                                           options: FileManager.DirectoryEnumerationOptions())
            for url in urls {
                try fileManager.removeItem(at: url)
            }
        } catch {
            print("Could not clear folder: \(error)")
        }
    }
    
    private func getOCRConfig(result: @escaping FlutterResult) {
        if let config = OCRConfigGetter.encodedOCRConfig {
            result(config)
        } else {
            result(OCRPerformerError.ocrConfig.value)
        }
    }
    
    private func detectDocument(_ parameters: CallParameters) {
        let result = parameters.result
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let page = SBSDKUIPage(from: arguments),
              let image = page.originalImage() else {
            self.returnArgumentsError(call: parameters.call, result: result)
            return
        }
        
        let polygon = DocumentDetector().detectPolygon(on: image)
        page.polygon = polygon
        
        result(page.encodedString)
    }
    
    private func performOCR(_ parameters: CallParameters) {
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let config = OCRPerformConfiguration(from: arguments) else {
            self.returnArgumentsError(call: parameters.call, result: parameters.result)
            return
        }
        OCRPerformer().perform(with: config, result: parameters.result)
    }
    
    private func refreshImageURLS(call: FlutterMethodCall, result: @escaping FlutterResult) {
        // As the location of the Application Support folder can change (e.g. on app update), we simply recreate
        // Page objects here to get a refreshed absolute file paths of the images.
        // (under the hood the native iOS SDK works with relative paths and is able to handle variable Application Support location)
        guard let arguments = call.arguments as? [String: Any]  else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let pages = SBSDKUIPage.pages(from: arguments)
        let dictionary = ["pages": pages.map({$0.encodedDictionary})]
        result(dictionary.encodedString)
    }
    
    private func estimateQuality(_ parameters: CallParameters) {
        let result = parameters.result
        DispatchQueue.global(qos: .background).async {
            guard let arguments = parameters.call.arguments as? [String: Any],
                  let page = SBSDKUIPage(from: arguments),
                  let image = page.documentImage() else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            let value = SBSDKDocumentQualityAnalyzer().analyze(on: image)
            
            DispatchQueue.main.async {
                result(value.encodedString)
            }
        }
    }
    
    private func createPDF(_ parameters: CallParameters) {
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let pageSize = SBSDKPDFRendererPageSize(from: arguments), 
              let pageOrientation = SBSDKPDFRendererPageOrientation(from: arguments) else {
            self.returnArgumentsError(call: parameters.call, result: parameters.result)
            return
        }
        var pdfAttributes: [String: Any]?
        if let options = arguments["options"] as? [String: Any], let metadata = options["pdfAttributes"] as? [String: Any] {
            pdfAttributes = metadata
        }
        let pages = SBSDKUIPage.pages(from: arguments)
        Renderer().renderPDF(from: pages,
                             pageSize: pageSize,
                             pageOrientation: pageOrientation,
                             pdfAttributes: pdfAttributes,
                             result: parameters.result)
    }
    
    private func createTIFF(_ parameters: CallParameters) {
        
        guard let arguments = parameters.call.arguments as? [String: Any]  else {
            self.returnArgumentsError(call: parameters.call, result: parameters.result)
            return
        }
        let options = TIFFOptions(from: arguments)
        let pages = SBSDKUIPage.pages(from: arguments)
        Renderer().renderTIFF(from: pages, options: options, result: parameters.result)
    }
    
    private func createPage(_ parameters: CallParameters) {
        let result = parameters.result
        
        DispatchQueue.global(qos: .background).async {
            guard let arguments = parameters.call.arguments as? [String: Any],
                  let imageURLString = arguments["originalImageFileUri"] as? String,
                  let shouldDetect = arguments["shouldDetectDocument"] as? Bool  else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            guard let imageURL = URL(string: imageURLString) else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            guard let imageData = try? Data(contentsOf: imageURL),
                  var image = UIImage(data: imageData)?.sbsdk_imageWithNormalizedOrientation() else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            if let documentImageSizeLimit = arguments["originalImageSizeLimit"] as? [String : Any] {
                let size = CGSize(from: documentImageSizeLimit)
                image = image.sbsdk_limited(to: size) ?? image
            }
            
            guard let uuid = SBSDKUIPageFileStorage.default().add(image) else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            var polygon: SBSDKPolygon?
            if shouldDetect {
                polygon = DocumentDetector().detectPolygon(on: image)
            }
            
            let page = SBSDKUIPage(pageFileID: uuid, polygon: polygon)
            if let documentImageSizeLimit = arguments["documentImageSizeLimit"] as? [String : Any] {
                page.documentImageSizeLimit = CGSize(from: documentImageSizeLimit)
            }
            
            DispatchQueue.main.async {
                result(page.encodedString)
            }
        }
        
    }
    
    private func createPageWithImageBytes(_ parameters: CallParameters) {
        let result = parameters.result
        
        DispatchQueue.global(qos: .background).async {
            guard
                let arguments = parameters.call.arguments as? [String: Any],
                let shouldDetect = arguments["shouldDetectDocument"] as? Bool
            else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            let imageBytes = arguments["originalImageBytes"] as? FlutterStandardTypedData
            guard let imageData = imageBytes?.data else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            guard var image = UIImage(data: imageData)?.sbsdk_imageWithNormalizedOrientation() else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            if let documentImageSizeLimit = arguments["originalImageSizeLimit"] as? [String : Any] {
                let size = CGSize(from: documentImageSizeLimit)
                image = image.sbsdk_limited(to: size) ?? image
            }
            
            guard let uuid = SBSDKUIPageFileStorage.default().add(image) else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            
            var polygon: SBSDKPolygon?
            if shouldDetect {
                polygon = DocumentDetector().detectPolygon(on: image)
            }
            
            let page = SBSDKUIPage(pageFileID: uuid, polygon: polygon)
            
            if let documentImageSizeLimit = arguments["documentImageSizeLimit"] as? [String : Any] {
                page.documentImageSizeLimit = CGSize(from: documentImageSizeLimit)
            }
            
            DispatchQueue.main.async {
                result(page.encodedString)
            }
        }
        
    }
    
    private func rotatePage(_ parameters: CallParameters) {
        let result = parameters.result
        DispatchQueue.global(qos: .background).async {
            
            guard let arguments = parameters.call.arguments as? [String: Any],
                  let page = SBSDKUIPage(from: arguments),
                  let count = arguments["rotationsCount"] as? Int else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            page.rotateClockwise(count)
            DispatchQueue.main.async {
                result(page.encodedString)
            }
        }
    }
    
    private func cropAndRotatePage(_ parameters: CallParameters) {
        let result = parameters.result
        DispatchQueue.global(qos: .background).async {
            
            guard let arguments = parameters.call.arguments as? [String: Any],
                  let page = SBSDKUIPage(from: arguments),
                  let count = arguments["rotationsCount"] as? Int,
                  let polygon = SBSDKPolygon(from: arguments)
            else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            page.rotateClockwise(count)
            page.polygon = polygon
            DispatchQueue.main.async {
                result(page.encodedString)
            }
        }
    }
    
    private func applyImageFilter(_ parameters: CallParameters) {
        let result = parameters.result
        
        DispatchQueue.global(qos: .background).async {
            
            guard let arguments = parameters.call.arguments as? [String: Any],
                  let page = SBSDKUIPage(from: arguments),
                  let filter = SBSDKImageFilterType(from: arguments) else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            page.filter = filter
            DispatchQueue.main.async {
                result(page.encodedString)
            }
        }
    }
    
    private func getFilteredDocumentPreviewURL(_ parameters: CallParameters) {
        let result = parameters.result
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let page = SBSDKUIPage(from: arguments),
              let filter = SBSDKImageFilterType(from: arguments) else {
            self.returnArgumentsError(call: parameters.call, result: result)
            return
        }
        let url = page.documentPreviewImageURL(using: filter)
        result(url?.absoluteString)
    }
    
    private func deletePage(_ parameters: CallParameters) {
        let result = parameters.result
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let page = SBSDKUIPage(from: arguments) else {
            self.returnArgumentsError(call: parameters.call, result: result)
            return
        }
        SBSDKUIPageFileStorage.default().removePageFileID(page.pageFileUUID)
        result(nil)
    }
    
    // detect barcodes from multiple images
    private func detectBarcodesOnImages(_ parameters: CallParameters) {
        let result = parameters.result
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let arrImageFilesUris = arguments["uris"] as? [String] else {
            self.returnArgumentsError(call: parameters.call, result: result)
            return
        }
        
        if !arrImageFilesUris.isEmpty {
            var arrResult: [[String: Any]] = []
            for imageUri in arrImageFilesUris {
                let url = URL(string: imageUri)?.path ?? ""
                if let image = UIImage(contentsOfFile: url) {
                    let barcodeTypes = SBSDKBarcodeType.fromArguments(arguments) ?? []
                    let acceptedDocumentFormats = SBSDKBarcodeDocumentType.fromArguments(arguments) ?? []
                    let detector = SBSDKBarcodeScanner()
                    detector.acceptedDocumentTypes = acceptedDocumentFormats
                    detector.acceptedBarcodeTypes = barcodeTypes
                    let barcodeResults = detector.detectBarCodes(on: image) ?? []
                    let barcodeList = barcodeResults.map({ $0.encodedDictionary })
                    
                    arrResult.append(barcodeList.isEmpty ?
                                     self.getBarcodeDetectionError(uri: imageUri, errorMessage: "No barcodes found.") :
                                        ["barcodeImageURI": imageUri, "barcodeItems": barcodeList])
                } else {
                    arrResult.append(self.getBarcodeDetectionError(uri: imageUri,
                                                                   errorMessage: "Unable to load given image."))
                }
            }
            result((["barcodeResults": arrResult] as [String: Any]).encodedString)
        }
    }
    
    // get the error message dictionary object
    private func getBarcodeDetectionError(uri: String, errorMessage: String) -> [String: Any] {
        return  ["barcodeImageURI": uri,
                 "error": errorMessage,
                 "barcodeItems": [] as [[String : Any?]]]
    }
}
